﻿module coneneko.window;
private import hdk.window;
private import hdk.control;
private import hdk.sdlttf;
private import std.cstream;
private import nanami.model;
private import nanami.effect;
private import coneneko.location;
private import std.random;
private import std.utf;
private import std.string;
private import std.file;
private import nanami.png;
private import std.thread;

interface _Window
{
	void main();
	void initialize();
	
	void text(dchar[] a);
	void waitInput();
	void backImage(dchar[] textureFileName);
	ubyte waitChoice(dchar[][] choices ...);
	void createModel(dchar[] name, dchar[] nanamiFileName);
	void deleteModel(dchar[] name);
	ModelBox[dchar[]] model();
	void playEffect(dchar[] fileName, uint divisionLength, Rectangle r);
}

auto class Window : hdk.window.Window, _Window
{
	private final View view;
	private ModelBox[dchar[]] modelMap;
	
	this()
	{
		view = new View(rootControl, new PngSaver(rootControl, this), this);
	}
	
	void main() {}
	
	void initialize()
	{
		view.initialize();
		modelMap = null;
		sleep(10);
	}
	
	void text(dchar[] a)
	{
		view.text = a;
	}
	
	void waitInput()
	{
		view.waitInput();
	}
	
	void backImage(dchar[] textureFileName)
	{
		view.backImage = textureFileName;
	}
	
	ubyte waitChoice(dchar[][] choices ...)
	{
		return view.waitChoice(choices);
	}
	
	void createModel(dchar[] name, dchar[] nanamiFileName)
	{
		if (name in modelMap) return;
		modelMap[name] = new ModelBox(rootControl, nanamiFileName);
		modelMap[name].z += 1.5;
	}
	
	void deleteModel(dchar[] name)
	{
		if (!(name in modelMap)) return;
		modelMap[name].cutLink();
		modelMap.remove(name);
	}
	
	ModelBox[dchar[]] model()
	{
		return modelMap;
	}
	
	void playEffect(dchar[] fileName, uint divisionLength, Rectangle r)
	{
		if (view.resetting) return;
		if (view.loadingTarget) return;
		EffectBox e = new EffectBox(rootControl, fileName, divisionLength, r.toMatrix());
		e.z += 2.0;
	}
}

class View : Control
{
	private final Background background;
	private final Control panel;
	private final TextArea textArea;
	private final SaveButton saveButton;
	private final LoadButton loadButton;
	private Location location, loadingTarget;
	private final PngSaver pngSaver;
	private Thread wnd;
	
	private bit waiting = false, skipping = false, resetting = false;
	
	this(Control parent, PngSaver pngSaver, Thread wnd)
	{
		super(parent);
		this.pngSaver = pngSaver;
		this.wnd = wnd;
		
		background = new Background(this);
		panel = new Control(this);
		textArea = new TextArea(panel);
		saveButton = new SaveButton(panel, &save);
		loadButton = new LoadButton(panel, &load);
	}
	
	void initialize()
	{
		background.textureFileName = null;
		textArea.text = null;
		skipping = false;
		location = new Location();
		resetting = false;
	}
	
	override bit inner(Point p)
	{
		return true;
	}
	
	override void input(uint id)
	{
		if (id == 4) skipping = true;
		else if (id == 5) skipping = false;
		else if (panel.visible && waiting && id == 0) waiting = false;
		else if (!panel.visible && id == 0) panel.visible = true;
		else if (id == 2) panel.visible = !panel.visible;
	}
	
	void text(dchar[] a)
	{
		textArea.text = a;
	}
	
	void waitInput()
	{
		if (resetting) return;
		if (loadingTarget)
		{
			if (location == loadingTarget)
			{
				loadingTarget = null;
				wnd.resume();
				new MessageBox(this, "loadしました");
			}
			else
			{
				location++;
				return;
			}
		}
		
		waiting = true;
		while (waiting)
		{
			if (!wr) throw new Error("waitInput");
			sleep(10);
			
			if (resetting) return;
			if (skipping)
			{
				waiting = false;
				break;
			}
		}
		
		location++;
	}
	
	void backImage(dchar[] textureFileName)
	{
		background.textureFileName = textureFileName;
	}
	
	ubyte waitChoice(dchar[][] choices ...)
	{
		if (resetting) return 0;
		if (loadingTarget)
		{
			if (location == loadingTarget)
			{
				loadingTarget = null;
				wnd.resume();
				new MessageBox(this, "loadしました");
			}
			else
			{
				uint result = Location.nextChoice(loadingTarget, location);
				location.addChoice(result);
				return result;
			}
		}
		if (skipping) skipping = false;
		
		ChoiceDialogBox dlg = new ChoiceDialogBox(panel, choices);
		dlg.waitInput(&resetting);
		dlg.cutLink();
		if (resetting) return 0;
		
		location.addChoice(dlg.result);
		return dlg.result;
	}
	
	private void save()
	{
		if (!exists("save")) mkdir("save");
		isdir("save");
		if (exists(saveFileName(99))) throw new Error("saveできない");
		
		uint saveIndex = uint.max;
		char[][] fileList = listdir("save");
		if (fileList.length < 1) saveIndex = 0;
		else
		{
			for (int i = 99; i >= 0; i--)
			{
				if (`save\` ~ fileList[$ - 2] == saveFileName(i))
				{
					saveIndex = i + 1;
					break;
				}
			}
		}
		if (saveIndex == uint.max) throw new Error("saveできない");
		
		write(saveFileName(saveIndex), location.toString());
		pngSaver.save(saveFileName(saveIndex) ~ ".png");
	}
	
	private void load(dchar[] fileName)
	{
		loadingTarget = new Location(cast(char[])read(toUTF8(fileName)));
		resetting = true;
		wnd.pause();
	}
	
	unittest
	{
		assert(`save\_0000` == saveFileName(0));
		assert(`save\_0001` == saveFileName(1));
		assert(`save\_0099` == saveFileName(99));
	}
	
	private static char[] saveFileName(uint index)
	{
		return format(`save\_%04d`, index);
	}
}

class Background : ImageBox
{
	this(Control parent)
	{
		super(parent);
		modelView = Rectangle(-1, -1, 2, 2).toMatrix;
		color = Vector(1, 1, 1, 1);
	}
}

class TextArea : Control
{
	private final Canvas backCanvas;
	private final StaticText[] staticTextList;
	
	this(Control control)
	{
		super(control);
		
		backCanvas = new Canvas(this);
		backCanvas.modelView = Rectangle(-0.7, -0.8, 1.6, 0.5).toMatrix();
		backCanvas.color = Vector(0, 0, 0, 0.5);
		
		for (int i = 0; i < 3; i++)
		{
			StaticText st = new StaticText(this);
			st.modelView = Rectangle(-0.6, -0.5 - 0.1 * i, 1.2, 0.1).toTextMatrix();
			st.color = Vector(1, 1, 1, 1.0);
			staticTextList ~= st;
		}
	}
	
	void text(dchar[] a)
	{
		if (!a) return;
		
		dchar[][] splitedTextList = splitTextTexture512x32(a);
		for (int i = 0; i < 3 && i < splitedTextList.length; i++)
		{
			staticTextList[i].text = splitedTextList[i];
		}
	}
}

class SaveButton : Button
{
	this(Control parent, void delegate() save)
	{
		super(parent);
		text = "save";
		rectangle = Rectangle(-0.9, -0.6, 0.2, 0.1);
		foreColor = Vector(1, 1, 1, 1);
		backColor = Vector(1, 0, 0, 1);
		pushedEvent = save;
	}
}

class LoadButton : Button
{
	private void delegate(dchar[] fileName) load;
	
	this(Control parent, void delegate(dchar[] fileName) load)
	{
		super(parent);
		this.load = load;
		
		text = "load";
		rectangle = Rectangle(-0.9, -0.8, 0.2, 0.1);
		foreColor = Vector(1, 1, 1, 1);
		backColor = Vector(0, 0, 1, 1);
		pushedEvent = &showFs;
	}
	
	void showFs()
	{
		new FileSelector(this, load);
	}
}

class ChoiceDialogBox : Control
{
	private final Canvas backCanvas;
	private final ButtonEx[] choicesButton;
	
	ubyte result = ubyte.max;
	
	this(Control parent, dchar[][] choices ...)
	{
		super(parent);
		z += 1.0;
		
		backCanvas = new Canvas(this);
		with (backCanvas)
		{
			modelView = Rectangle(-0.5, -0.2, 1.0, 1.0).toMatrix();
			backCanvas.color = Vector(1, 1, 1, 0.1);
		}
		
		foreach (int i, dchar[] a; choices)
		{
			choicesButton ~= new ButtonEx(backCanvas, a, i, choices.length, &result);
		}
	}
	
	void waitInput(bit* resetting)
	{
		while (result == ubyte.max)
		{
			if (!wr) throw new Error("waitInput");
			if (*resetting) return;
			sleep(10);
		}
	}
	
	private class ButtonEx : Button
	{
		private ubyte* result;
		private final ubyte index;
		
		this(Control parent, dchar[] text, ubyte index, ubyte totalLength, ubyte* result)
		{
			super(parent);
			this.result = result;
			this.index = index;
			
			this.text = text;
			float h = 1.2 / (totalLength + 1);
			rectangle = Rectangle(-0.5, 0.8 - h * (index + 1), 1.0, 0.1);
			backColor = Vector(1, 1, 1, 1);
			foreColor = Vector(1, 1, 1, 1);
			pushedEvent = &push;
		}
		
		void push() { *result = index; }
	}
}

class ModelBox : Control
{
	private Model model;
	private Morph[] morphList;
	
	uint nextMotion = uint.max;
	
	this(Control parent, dchar[] nanamiFileName)
	{
		model = new Model(nanamiFileName);
		super(parent);
		morphList.length = model.height < 2 ? 0 : model.height - 2;
	}
	
	void position(Vector p)
	{
		model.modelView = Matrix.translation(p)
			* Matrix.lookAt(0.0, 40.0, 150.0,  0.0, 40.0, 0.0);
	}
	
	void motion(uint a)
	{
		nextMotion = uint.max;
		model.motionIterator.y = a;
		model.motionIterator.x = 0;
	}
	
	override void draw()
	{
		tickMotion();
		tickMorph();
		model(wr);
	}
	
	private void tickMotion()
	{
		MotionIterator i = model.motionIterator;
		i++;
		if (!i.isRangeOut) return;
		
		if (nextMotion != uint.max)
		{
			i.y = nextMotion;
			nextMotion = uint.max;
		}
		i.x = 0;
	}
	
	void visibleItem(uint index, bit visible)
	{
		model[1, index].visible = visible;
	}
	
	void playMorphing(uint index, float frequencyPerSecond, float span)
	{
		morphList[index] = new Morph(index, frequencyPerSecond, span);
	}
	
	void stopMorphing(uint index)
	{
		if (morphList[index]) delete morphList[index];
	}
	
	private void tickMorph()
	{
		foreach (Morph a; morphList) if (a) a.tick();
	}
	
	private class Morph
	{
		uint y, iterator = 0;
		float frequencyPerFrame, span;
		bit playing = false;
		const float FRAME_LENGTH = 1000.0 / 60.0;
		float t = 0.0;
		
		this(uint y, float frequencyPerSecond, float span)
		{
			this.y = y + 2;
			this.frequencyPerFrame = frequencyPerSecond / FRAME_LENGTH;
			this.span = span;
		}
		
		~this()
		{
			model[y, iterator].visible = false;
		}
		
		void tick()
		{
			if (!playing && frand() < frequencyPerFrame)
			{
				playing = true;
				t = 0.0;
			}
			
			if (!playing) return;
			
			model[y, iterator].visible = false;
			
			t += 1.0 / 60.0;
			
			iterator = computeIterator();
			if (iterator >= model.width(y))
			{
				playing = false;
				iterator = 0;
			}
			
			model[y, iterator].visible = true;
		}
		
		private float frand()
		out (result)
		{
			assert(0.0 <= result && result <= 1.0);
		}
		body
		{
			return cast(float)rand() / cast(float)uint.max;
		}
		
		private uint computeIterator()
		{
			return cast(uint)(cast(float)model.width(y) * (t / span));
		}
	}
}

class EffectBox : Control
{
	private Effect effect;
	
	this(Control parent, dchar[] fileName, uint divisionLength, Matrix modelView)
	{
		effect = new Effect(fileName, divisionLength);
		effect.modelView = modelView;
		super(parent);
	}
	
	override void draw()
	{
		if (!effect) return;
		
		if (effect.alive)
		{
			effect(wr);
		}
		else
		{
			cutLink();
			delete effect;
		}
	}
}

class MessageBox : Canvas
{
	private StaticText staticText;
	private uint t = 0;
	
	this(Control parent, dchar[] msg)
	{
		super(parent);
		modelView = Rectangle(-1.0, -0.05, 2.0, 0.1).toMatrix();
		color = Vector(0, 1, 0, 0.4);
		
		staticText = new StaticText(this);
		staticText.modelView = Rectangle(-0.5, -0.05, 1.0, 0.1).toTextMatrix();
		staticText.text = msg;
		staticText.color = Vector(1, 1, 1, 1);
	}
	
	override void draw()
	{
		t++;
		if (t > 40) cutLink();
		super.draw();
	}
}

class FileSelector : Canvas
{
	private final ImageButton[] imageButtonList;
	private final Button upButton, downButton, up5Button, down5Button;
	private final uint lastIndex;
	private uint currentIndex;
	
	this(Control parent, void delegate(dchar[] fileName) load)
	{
		super(parent);
		
		currentIndex = lastIndex = computeLastIndex();
		
		z += 4.0;
		modelView = Matrix.scaling(2, 2, 1);
		color = Vector(0, 0, 0.4, 0.8);
		
		for (int i = 0; i < 5; i++)
		{
			ImageButton ib = new ImageButtonEx(this, load);
			with (ib)
			{
				rectangle = Rectangle(-0.5, 0.65 - 0.4 * i, 1.0, 0.3);
				text = toUTF32(.toString(i));
				foreColor = Vector(1, 1, 1, 1);
				backColor = Vector(1, 1, 1, 0.5);
				imageWidth = 640.0 * 0.3 / 480.0;
				textHeight = 0.1;
			}
			imageButtonList ~= ib;
		}
		
		up5Button = new Button(this);
		with (up5Button)
		{
			text = "↑↑";
			rectangle = Rectangle(0.6, 0.75, 0.3, 0.15);
			foreColor = Vector(1, 1, 1, 1);
			backColor = Vector(1, 1, 1, 1);
			pushedEvent = &up5;
		}
		
		upButton = new Button(this);
		with (upButton)
		{
			text = "↑";
			rectangle = Rectangle(0.6, 0.35, 0.3, 0.3);
			foreColor = Vector(1, 1, 1, 1);
			backColor = Vector(1, 1, 1, 1);
			pushedEvent = &up;
		}
		
		downButton = new Button(this);
		with (downButton)
		{
			text = "↓";
			rectangle = Rectangle(0.6, -0.65, 0.3, 0.3);
			foreColor = Vector(1, 1, 1, 1);
			backColor = Vector(1, 1, 1, 1);
			pushedEvent = &down;
		}
		
		down5Button = new Button(this);
		with (down5Button)
		{
			text = "↓↓";
			rectangle = Rectangle(0.6, -0.9, 0.3, 0.15);
			foreColor = Vector(1, 1, 1, 1);
			backColor = Vector(1, 1, 1, 1);
			pushedEvent = &down5;
		}
		
		update();
	}
	
	override bit inner(Point p)
	{
		return true;
	}
	
	override void input(uint id)
	{
		if (id == 2) cutLink();
	}
	
	override void draw()
	{
		super.draw();
	}
	
	private static uint computeLastIndex()
	{
		if (!exists("save"))
		{
			mkdir("save");
			return 0;
		}
		isdir("save");
		char[][] fileList = listdir("save");
		if (fileList.length < 1) return 0;
		
		for (int i = 99; i >= 0; i--)
		{
			if (`save\` ~ fileList[$ - 2] == View.saveFileName(i)) return i;
		}
		
		throw new Error("loadできない");
	}
	
	unittest
	{
		assert(0 == viewIndex(0));
		assert(0 == viewIndex(4));
		assert(1 == viewIndex(5));
		assert(19 == viewIndex(99));
	}
	
	private static uint viewIndex(uint index)
	{
		return index / 5;
	}
	
	private void up5() { currentIndex -= 25; update(); }
	private void up() { currentIndex -= 5; update(); }
	private void down() { currentIndex += 5; update(); }
	private void down5() { currentIndex += 25; update(); }
	
	private void update()
	{
		assert(lastIndex < 100);
		assert(0 <= currentIndex && currentIndex <= lastIndex);
		
		up5Button.visible = currentIndex >= 25 ? true : false;
		upButton.visible = currentIndex >= 5 ? true : false;
		downButton.visible = currentIndex + 5 <= lastIndex ? true : false;
		down5Button.visible = currentIndex + 25 <= lastIndex ? true : false;
		
		for (int i = 0; i < 5; i++)
		{
			uint index = viewIndex(currentIndex) * 5 + i;
			
			char[] sfn = View.saveFileName(index);
			sfn = exists(sfn) ? sfn : null;
			
			char[] pngFileName = View.saveFileName(index) ~ ".png";
			pngFileName = exists(pngFileName) ? pngFileName : null;
			
			if (sfn && pngFileName)
			{
				imageButtonList[i].text = toUTF32(sfn);
				imageButtonList[i].textureFileName = toUTF32(pngFileName);
			}
			else
			{
				imageButtonList[i].text = null;
				imageButtonList[i].textureFileName = null;
			}
		}
	}
	
	class ImageButtonEx : ImageButton
	{
		private final void delegate(dchar[] fileName) loadFile;
		
		this(Control parent, void delegate(dchar[] fileName) loadFile)
		{
			super(parent);
			this.loadFile = loadFile;
			pushedEvent = &load;
		}
		
		void load()
		{
			if (!text) return;
			loadFile(text);
			parent.cutLink();
		}
	}
}

class PngSaver : Control
{
	private final uint width, height;
	private uint[] delegate() readBackBuffer;
	private char[] saveFileName;
	
	this(Control parent, hdk.window.Window wnd)
	{
		super(parent);
		z += 20.0;
		readBackBuffer = &wnd.base.readBackBuffer;
		width = wnd.width;
		height = wnd.height;
	}
	
	override void draw()
	{
		if (!saveFileName) return;
		uint[] image = new uint[256 * 256];
		uint[] buffer = readBackBuffer();
		for (int y = 0; y < 256; y++)
		{
			for (int x = 0; x < 256; x++)
			{
				image[256 * y + x] =
					buffer[width * (y * height / 256) + (x * width / 256)];
			}
		}
		writePng(saveFileName, image, 256, 256);
		delete image;
		delete buffer;
		new MessageBox(this, toUTF32(saveFileName[0..$ - 4]) ~ "にsaveしました");
		saveFileName = null;
	}
	
	void save(char[] fileName)
	{
		saveFileName = fileName;
	}
}
