﻿module hdk.oglvertexbuffer;
private import sdl;
private import hdk.openglarb;
private import std.random;
private import std.utf;

interface OglVertexBufferMap
{
	bit hasVertexBuffer(dchar[] key);
	void addPcntVertexBuffer(dchar[] key, void[] pcntBuffer, uint vertexLength);
	void addPcntmwVertexBuffer(dchar[] key, void[] pcntmwBuffer, uint vertexLength);
	uint getVertexBufferByteLength(dchar[] key);
	void[] getVertexBufferData(dchar[] key);
	void attachVertexBuffer(dchar[] key);
	void attachVertexBuffer(dchar[] key, uint matrixIndexLocation, uint weightLocation);
	void detachVertexBuffer();
	uint getAttachingVertexBufferLength();
	void clearVertexBufferMap();
	uint totalVertexBufferSize();
}

// position, color, normal, texCoord, matrixIndex, weightの順
// float3 * vertexLength
//  + float4 * vertexLength
//  + float3 * vertexLength
//  + float2 * vertexLength
//   + float2 * vertexLength
//   + float * weight

private class OglPcntVertexBuffer
{
	private import hdk.error;
	mixin OglError;
	
	protected OglVertexBuffer[] vertexBuffer;
	protected uint _vertexLength;
	uint getVertexLength() { return _vertexLength; }
	
	protected uint byteLengthPerVertex() { return float.sizeof * (3 + 4 + 3 + 2); }
	
	this(void[] pcnt, uint vertexLength)
	{
		static assert(ubyte.sizeof == void.sizeof);
		if (vertexLength % 3 != 0) error(__LINE__, "vertexLength % 3 == 0");
		if (pcnt.length / vertexLength != byteLengthPerVertex) error(__LINE__, "this");
		
		uint begin = 0, end = 0;
		
		begin = end;
		end += (float.sizeof * 3) * vertexLength;
		void[] positionBuffer = pcnt[begin..end];
		
		begin = end;
		end += (float.sizeof * 4) * vertexLength;
		void[] colorBuffer = pcnt[begin..end];
		
		begin = end;
		end += (float.sizeof * 3) * vertexLength;
		void[] normalBuffer = pcnt[begin..end];
		
		begin = end;
		end += (float.sizeof * 2) * vertexLength;
		void[] texCoordBuffer = pcnt[begin..end];
		
		vertexBuffer ~= new OglPositionBuffer(positionBuffer);
		vertexBuffer ~= new OglColorBuffer(colorBuffer);
		vertexBuffer ~= new OglNormalBuffer(normalBuffer);
		vertexBuffer ~= new OglTexCoordBuffer(texCoordBuffer);
		
		_vertexLength = byteLength / byteLengthPerVertex;
	}
	
	~this()
	{
		foreach (OglVertexBuffer a; vertexBuffer) delete a;
	}
	
	void attach()
	{
		foreach (OglVertexBuffer a; vertexBuffer) a.attach();
	}
	
	void detach()
	{
		foreach (OglVertexBuffer a; vertexBuffer) a.detach();
	}
	
	uint byteLength()
	{
		uint result = 0;
		foreach (OglVertexBuffer a; vertexBuffer) result += a.byteLength;
		return result;
	}
	
	void[] getBufferData()
	{
		void[] result;
		foreach (OglVertexBuffer a; vertexBuffer) result ~= a.getBufferData();
		return result;
	}
}

private class OglPcntmwVertexBuffer : OglPcntVertexBuffer
{
	private OglMatrixIndexBuffer miBuffer;
	private OglWeightBuffer wBuffer;
	
	this(void[] pcntmw, uint vertexLength)
	{
		super(pcntmw, vertexLength);
		uint begin = 0, end = 0;
		begin = end = float.sizeof * 12 * vertexLength;
		end += (float.sizeof * 2) * vertexLength;
		void[] matrixIndexBuffer = pcntmw[begin..end];
		
		begin = end;
		end += (float.sizeof * 1) * vertexLength;
		void[] weightBuffer = pcntmw[begin..end];
		
		miBuffer = new OglMatrixIndexBuffer(matrixIndexBuffer);
		wBuffer = new OglWeightBuffer(weightBuffer);
		vertexBuffer ~= miBuffer;
		vertexBuffer ~= wBuffer;
		
		_vertexLength = byteLength / byteLengthPerVertex;
	}
	
	protected override uint byteLengthPerVertex()
	{
		return float.sizeof * (3 + 4 + 3 + 2 + 2 + 1);
	}
	
	void attach(uint matrixIndexLocation, uint weightLocation)
	{
		miBuffer.setLocation(matrixIndexLocation);
		wBuffer.setLocation(weightLocation);
		super.attach();
	}
}

private class OglVertexBuffer
{
	private import hdk.error;
	mixin OglError;
	
	private uint vertexBuffer;
	private bit lock = false;
	
	this(void[] data)
	{
		glGenBuffersARB(1, &vertexBuffer);
		bind();
		glBufferDataARB(GL_ARRAY_BUFFER_ARB,
			void.sizeof * data.length, data.ptr, GL_STATIC_DRAW_ARB);
	}
	
	~this()
	{
		if (lock) detach();
		glDeleteBuffersARB(1, &vertexBuffer);
	}
	
	private void bind() { glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexBuffer); }
	
	uint byteLength()
	{
		throwIfLockError();
		int result;
		bind();
		glGetBufferParameterivARB(GL_ARRAY_BUFFER_ARB, GL_BUFFER_SIZE_ARB, &result);
		return result;
	}
	
	void[] getBufferData()
	{
		throwIfLockError();
		ubyte[] result = new ubyte[byteLength];
		bind();
		glGetBufferSubDataARB(GL_ARRAY_BUFFER_ARB, 0,
			cast(GLsizeiptrARB)result.length, result.ptr);
		return result;
	}
	
	void attach()
	{
		throwIfLockError();
		lock = true;
		bind();
	}
	
	void detach() { lock = false; }
	
	private void throwIfLockError() { if (lock) throw new Error("lock is true"); }
}

private class OglMatrixIndexBuffer : OglVertexBuffer
{
	protected uint floatLength = 2;
	
	this(void[] data) { super(data); }
	
	private uint location;
	void setLocation(uint a) { location = a; }
	
	override void attach()
	{
		super.attach();
		glEnableVertexAttribArrayARB(location);
		glVertexAttribPointerARB(location, floatLength, GL_FLOAT, false, 0, null);
	}
	
	override void detach()
	{
		glDisableVertexAttribArrayARB(location);
		super.detach();
	}
}

private class OglWeightBuffer : OglMatrixIndexBuffer
{
	this(void[] data) { floatLength = 1; super(data); }
}

private class OglPositionBuffer : OglVertexBuffer
{
	this(void[] data) { super(data); }
	
	override void attach()
	{
		super.attach();
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, null);
	}
	
	override void detach()
	{
		glDisableClientState(GL_VERTEX_ARRAY);
		super.detach();
	}
}

private class OglColorBuffer : OglVertexBuffer
{
	this(void[] data) { super(data); }
	
	override void attach()
	{
		super.attach();
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4, GL_FLOAT, 0, null);
	}
	
	override void detach()
	{
		glDisableClientState(GL_COLOR_ARRAY);
		super.detach();
	}
}

private class OglNormalBuffer : OglVertexBuffer
{
	this(void[] data) { super(data); }
	
	override void attach()
	{
		super.attach();
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT, 0, null);
	}
	
	override void detach()
	{
		glDisableClientState(GL_NORMAL_ARRAY);
		super.detach();
	}
}

private class OglTexCoordBuffer : OglVertexBuffer
{
	this(void[] data) { super(data); }
	
	override void attach()
	{
		super.attach();
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		//glActiveTextureARB(GL_TEXTURE0_ARB); // 必要?
		//glClientActiveTextureARB(GL_TEXTURE0_ARB); // こっちは必要だと思う
		glTexCoordPointer(2, GL_FLOAT, 0, null);
	}
	
	override void detach()
	{
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		super.detach();
	}
}

template OglVertexBufferMap_()
{
	private import hdk.error;
	mixin OglError;
	
	private static bit initialized = false;
	
	private OglPcntVertexBuffer[dchar[]] pcntMap;
	private OglPcntmwVertexBuffer[dchar[]] pcntmwMap;
	private OglPcntVertexBuffer lockItem1;
	private OglPcntmwVertexBuffer lockItem2;
	private bit lock() { return lockItem1 || lockItem2 ? true : false; }
	
	this()
	{
		if (initialized) error(__LINE__, "OglVertexBufferMap_.this");
		initialized = true;
		initializeArb();
	}
	
	~this()
	{
		clearVertexBufferMap();
		initialized = false;
	}
	
	void clearVertexBufferMap()
	{
		if (lock) detachVertexBuffer();
		foreach (OglPcntVertexBuffer a; pcntMap) delete a;
		foreach (OglPcntmwVertexBuffer a; pcntmwMap) delete a;
		pcntMap = null;
		pcntmwMap = null;
		totalSize = 0;
	}
	
	bit hasVertexBuffer(dchar[] key)
	{
		throwIfLockError();
		return (key in pcntMap) || (key in pcntmwMap) ? true : false;
	}
	
	void addPcntVertexBuffer(dchar[] key, void[] pcntBuffer, uint vertexLength)
	{
		throwIfLockError();
		if (hasVertexBuffer(key)) error(__LINE__, "addPcnt");
		pcntMap[key] = new OglPcntVertexBuffer(pcntBuffer, vertexLength);
		totalSize += float.sizeof * (3 + 4 + 3 + 2) * vertexLength;
	}
	
	void addPcntmwVertexBuffer(dchar[] key, void[] pcntmwBuffer, uint vertexLength)
	{
		throwIfLockError();
		if (hasVertexBuffer(key)) error(__LINE__, "addPcntmw");
		pcntmwMap[key] = new OglPcntmwVertexBuffer(pcntmwBuffer, vertexLength);
		totalSize += float.sizeof * (3 + 4 + 3 + 2 + 2 + 1) * vertexLength;
	}
	
	uint getVertexBufferByteLength(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		return key in pcntMap ? pcntMap[key].byteLength : pcntmwMap[key].byteLength;
	}
	
	void[] getVertexBufferData(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		return key in pcntMap ? pcntMap[key].getBufferData() : pcntmwMap[key].getBufferData();
	}
	
	void attachVertexBuffer(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		if (!(key in pcntMap))
		{
			error(__LINE__, format(`not found "%s" in pcnt map`, toUTF8(key)));
		}
		lockItem1 = pcntMap[key];
		lockItem1.attach();
	}
	
	void attachVertexBuffer(dchar[] key, uint matrixIndexLocation, uint weightLocation)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		if (!(key in pcntmwMap)) error(__LINE__, "attach");
		lockItem2 = pcntmwMap[key];
		lockItem2.attach(matrixIndexLocation, weightLocation);
	}
	
	void detachVertexBuffer()
	{
		if (lockItem1) lockItem1.detach();
		else lockItem2.detach();
		lockItem1 = lockItem2 = null;
	}
	
	private void throwIfLockError()
	{
		if (lock) throw new Error("lock is true");
	}
	
	private void throwIfNotFoundError(dchar[] key)
	{
		if (!hasVertexBuffer(key)) error(__LINE__, format("%s not found", key));
	}
	
	uint getAttachingVertexBufferLength()
	{
		if (!lockItem1 && !lockItem2) error(__LINE__, "getAttachingVertexBufferLength");
		return lockItem1 ?
			lockItem1.getVertexLength() : lockItem2.getVertexLength();
	}
	
	private uint totalSize = 0;
	uint totalVertexBufferSize() { return totalSize; }
}

auto class OglVertexBufferMapTest : OglVertexBufferMap
{
	private import std.cstream;
	mixin OglVertexBufferMap_;
	
	static void execute()
	{
		try
		{
			initialize();
			initializeArb();
			testOglVertexBuffer();
			testOglPcntVertexBuffer();
			testOglPcntmwVertexBuffer();
			testOglVertexBufferMap();
		}
		catch (Exception e)
		{
			derr.writeLine("Exception:" ~ e.toString());
		}
		finally
		{
			SDL_Quit();
		}
	}
	
	private static void initialize()
	{
		const uint WIDTH = 1, HEIGHT = 1;
		SDL_Init(SDL_INIT_VIDEO);
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
		SDL_SetVideoMode(WIDTH, HEIGHT, 24, SDL_OPENGL);
	}
	
	static void testOglVertexBuffer()
	{
		ubyte[] testData = new ubyte[256];
		foreach (inout ubyte a; testData) a = cast(ubyte)rand();
		auto OglVertexBuffer vb = new OglVertexBuffer(testData);
		assert(ubyte.sizeof * testData.length == vb.byteLength);
		ubyte[] testData2 = cast(ubyte[])vb.getBufferData();
		assert(testData2 == testData);
	}
	
	static void testOglPcntVertexBuffer()
	{
		float[] testData = new float[12 * 30];
		foreach (inout float a; testData) a = cast(float)rand();
		auto OglPcntVertexBuffer vb = new OglPcntVertexBuffer(testData, 30);
		assert(float.sizeof * testData.length == vb.byteLength);
		assert(testData == vb.getBufferData());
		assert(GL_FALSE == glIsEnabled(GL_VERTEX_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_COLOR_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_NORMAL_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_TEXTURE_COORD_ARRAY));
		vb.attach();
		assert(GL_TRUE == glIsEnabled(GL_VERTEX_ARRAY));
		assert(GL_TRUE == glIsEnabled(GL_COLOR_ARRAY));
		assert(GL_TRUE == glIsEnabled(GL_NORMAL_ARRAY));
		assert(GL_TRUE == glIsEnabled(GL_TEXTURE_COORD_ARRAY));
		assert(30 == vb.getVertexLength());
		vb.detach();
		assert(GL_FALSE == glIsEnabled(GL_VERTEX_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_COLOR_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_NORMAL_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_TEXTURE_COORD_ARRAY));
	}
	
	static void testOglPcntmwVertexBuffer()
	{
		float[] testData = new float[15 * 30];
		foreach (inout float a; testData) a = cast(float)rand();
		auto OglPcntmwVertexBuffer vb = new OglPcntmwVertexBuffer(testData, 30);
		assert(float.sizeof * testData.length == vb.byteLength);
		assert(testData == vb.getBufferData());
		assert(GL_FALSE == glIsEnabled(GL_VERTEX_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_COLOR_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_NORMAL_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_TEXTURE_COORD_ARRAY));
		vb.attach(0, 1);
		assert(GL_TRUE == glIsEnabled(GL_VERTEX_ARRAY));
		assert(GL_TRUE == glIsEnabled(GL_COLOR_ARRAY));
		assert(GL_TRUE == glIsEnabled(GL_NORMAL_ARRAY));
		assert(GL_TRUE == glIsEnabled(GL_TEXTURE_COORD_ARRAY));
		assert(30 == vb.getVertexLength());
		vb.detach();
		assert(GL_FALSE == glIsEnabled(GL_VERTEX_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_COLOR_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_NORMAL_ARRAY));
		assert(GL_FALSE == glIsEnabled(GL_TEXTURE_COORD_ARRAY));
	}
	
	static void testOglVertexBufferMap()
	{
		const uint VERTEX_LENGTH = 40 * 3;
		float[] pcntTestData = new float[12 * VERTEX_LENGTH];
		float[] pcntmwTestData = new float[15 * VERTEX_LENGTH];
		foreach (inout float a; pcntTestData) a = cast(float)rand();
		foreach (inout float a; pcntmwTestData) a = cast(float)rand();
		
		auto OglVertexBufferMapTest test = new OglVertexBufferMapTest();
		with (test)
		{
			assert(!hasVertexBuffer("hoge"));
			addPcntVertexBuffer("hoge", pcntTestData, VERTEX_LENGTH);
			assert(hasVertexBuffer("hoge"));
			assert(float.sizeof * pcntTestData.length == getVertexBufferByteLength("hoge"));
			assert(pcntTestData == getVertexBufferData("hoge"));
			
			assert(!hasVertexBuffer("hoge2"));
			addPcntmwVertexBuffer("hoge2", pcntmwTestData, VERTEX_LENGTH);
			assert(hasVertexBuffer("hoge2"));
			assert(float.sizeof * pcntmwTestData.length == getVertexBufferByteLength("hoge2"));
			assert(pcntmwTestData == getVertexBufferData("hoge2"));
			
			attachVertexBuffer("hoge");
			assert(VERTEX_LENGTH == getAttachingVertexBufferLength());
			detachVertexBuffer();
			attachVertexBuffer("hoge2", 1, 2);
			assert(VERTEX_LENGTH == getAttachingVertexBufferLength());
			detachVertexBuffer();
		}
	}
}
