﻿module hdk.sdlimage;
version (hdk_lib) private import sdl_image;
private import std.file;
private import std.string;
private import std.stream;
private import std.utf;

/// 32bit rgba
uint[] readImage(dchar[] fileName, out uint width, out uint height)
{
	return SdlImage.readImage(toUTF8(fileName), width, height);
}

/// 書き出す
void writeBitmap(dchar[] fileName, uint[] rgbaPixels, uint width, uint height)
{
	SdlImage.writeBmpImage(toUTF8(fileName), rgbaPixels, width, height);
}


version (hdk_lib) {

private interface _SdlImage
{
	static uint[] readImage(char[] fileName, out uint width, out uint height);
	static uint[] readRgba32Pixels(SDL_Surface* surface);
	static void writeBmpImage(char[] fileName, void[] rgba, uint width, uint height);
}

private class SdlImage : _SdlImage
{
	private import hdk.error;
	mixin SdlError;
	
	static uint[] readImage(char[] fileName, out uint width, out uint height)
	{
		SDL_Surface* image = IMG_Load(toMBSz(fileName));
		if (!image) error(__LINE__, `read:"` ~ fileName ~ `"`);
		width = image.w;
		height = image.h;
		uint[] result = readRgba32Pixels(image);
		SDL_FreeSurface(image);
		return result;	
	}
	
	protected static uint[] readRgba32Pixels(SDL_Surface* surface)
	{
		SDL_PixelFormat format = getRgbaFormat();
		surface = SDL_ConvertSurface(surface, &format, SDL_SWSURFACE);
		uint[] result = new uint[surface.w * surface.h];
		memcpy(result.ptr, surface.pixels, uint.sizeof * result.length);
		return result;
	}
	
	private static SDL_PixelFormat getRgbaFormat()
	{
		SDL_PixelFormat result;
		with (result)
		{
			palette = null;
			BitsPerPixel = 32;
			BytesPerPixel = 4;
			Rmask = 0xff;
			Gmask = 0xff00;
			Bmask = 0xff0000;
			Amask = 0xff000000;
			Rshift = 0;
			Gshift = 8;
			Bshift = 16;
			Ashift = 24;
			Rloss = Gloss = Bloss = Aloss = 0;
			colorkey = 0;
			alpha = 255;
		}
		return result;
	}
	
	
	static void writeBmpImage(char[] fileName, void[] rgba, uint width, uint height)
	{
		if (rgba.length != uint.sizeof * width * height
			|| rgba == null) error(__LINE__, "writeBmpImage");
		
		BufferedFile result = new BufferedFile();
		result.create(fileName);
		writeHeader(result, width, height);
		
		uint[] srcBeginIndexList = new uint[height];
		for (int i = 0; i < srcBeginIndexList.length; i++)
		{
			srcBeginIndexList[i] = rgba.length - (uint.sizeof * width) * (i + 1);
		}
		
		for (int y = 0; y < height; y++)
		{
			int srcIndex = srcBeginIndexList[y];
			int writeLength = 0;
			for (int x = 0; x < width; x++)
			{
				ubyte[] src = cast(ubyte[])rgba;
				ubyte r = src[srcIndex++];
				ubyte g = src[srcIndex++];
				ubyte b = src[srcIndex++];
				srcIndex++;
				result.write(b);
				result.write(g);
				result.write(r);
				writeLength += 3;
			}
			int rowTailLength = bmpRowByteLength(width) - writeLength;
			assert(0 <= rowTailLength);
			for (int i = 0; i < rowTailLength; i++) result.write(cast(ubyte)0);
		}
		
		result.close();
	}
	
	private static void writeHeader(Stream s, uint width, uint height)
	{
		uint imageSize = bmpRowByteLength(width) * height;
		s.write('B');
		s.write('M');
		s.write(imageSize + 0x36);
		s.write(cast(uint)0);
		s.write(cast(uint)0x36); // Header Size
		s.write(cast(uint)0x28); // Info Header Size
		s.write(width);
		s.write(height);
		s.write(cast(ushort)1);
		s.write(cast(ushort)0x18); // 24bit
		s.write(cast(uint)0);
		s.write(cast(uint)(imageSize));
		s.write(cast(uint)0);
		s.write(cast(uint)0);
		s.write(cast(uint)0);
		s.write(cast(uint)0);
	}
	
	private static uint bmpRowByteLength(uint width) // 4の倍数のバイトサイズ
	out (result)
	{
		assert(0 == result % 4);
	}
	body
	{
		uint result = 3 * width;
		if (result % 4 != 0) result += 4 - result % 4;
		return result;
	}
}

} // version (hdk_lib) {
