module nanami.anchor;
private import std.cstream;
private import std.string;
private import hdk.math;
private import nanami.mqo;


interface _Anchor
{
	static Anchor[] createList(Mqo.Material[] material, Mqo.Object anchorObject);
	//Anchor.this(Vector[] triangleList, char[] materialName)
	bit has(Vector point);
	char[] materialName();
	static bit isAnchor(Mqo.Object obj);
	static char[] extractRelationalObjectName(Mqo.Object anchorObject);
}


class Anchor : _Anchor
{
	unittest
	{
		assert(isAnchor("anchor1|hoge"));
		assert(isAnchor("anchor2|hoge"));
		assert(!isAnchor("hoge"));
		Vector[] vertex = new Vector[8];
		vertex[0] = Vector.create(-100.0000, 100.0000, -100.0000);
		vertex[1] = Vector.create(100.0000, 100.0000, -100.0000);
		vertex[2] = Vector.create(-100.0000, 100.0000, 100.0000);
		vertex[3] = Vector.create(100.0000, 100.0000, 100.0000);
		vertex[4] = Vector.create(-100.0000, -100.0000, 100.0000);
		vertex[5] = Vector.create(100.0000, -100.0000, 100.0000);
		vertex[6] = Vector.create(-100.0000, -100.0000, -100.0000);
		vertex[7] = Vector.create(100.0000, -100.0000, -100.0000);
		const uint[] indexList =
		[
			1, 3, 2,  1, 2, 0,  3, 5, 4,  3, 4, 2,
			5, 7, 6,  5, 6, 4,  7, 1, 0,  7, 0, 6,
			7, 5, 3,  7, 3, 1,  0, 2, 4,  0, 4, 6,
		];
		Vector[] boxData = new Vector[indexList.length];
		for (int i = 0; i < indexList.length; i++)
		{
			boxData[i] = vertex[indexList[i]];
		}
		Anchor box = new Anchor(boxData, "hoge");
		assert(box.has(Vector.create(0, 0, 0)));
		assert(!box.has(Vector.create(0, 200, 0)));
		assert(!box.has(Vector.create(0, -200, 0)));
		assert(!box.has(Vector.create(150, 0, 0)));
		
		Vector[] polyData = new Vector[3];
		polyData[0] = Vector.create(0, 1, 0);
		polyData[1] = Vector.create(1, 0, 0);
		polyData[2] = Vector.create(0, 0, 1);
		Anchor poly = new Anchor(polyData, "hoge");
		assert(!poly.has(Vector.create(0.5, 0.5, 0.5)));
	}
	
	private final Vector[] triangleList;
	private final char[] _materialName;
	
	this(Vector[] triangleList, char[] materialName)
	in
	{
		assert(0 == triangleList.length % 3);
	}
	body
	{
		this.triangleList = triangleList;
		this._materialName = materialName;
	}
	
	char[] materialName() { return _materialName; }
	
	static bit isAnchor(Mqo.Object obj) { return isAnchor(obj.name); }
	private static bit isAnchor(char[] name)
	{
		return 6 <= name.length && name[0..6] == "anchor" ? true : false;
	}
	
	bit has(Vector point)
	{
		Vector[] dir = new Vector[6];
		dir[0] = Vector.create(1, 0, 0);
		dir[1] = Vector.create(0, 1, 0);
		dir[2] = Vector.create(0, 0, 1);
		bit[] bits = new bit[6];
		assert(false == bits[0]);
		for (int i = 0; i < triangleList.length; i += 3)
		{
			Vector p0 = triangleList[i];
			Vector p1 = triangleList[i + 1];
			Vector p2 = triangleList[i + 2];
			float dist;
			if (triangleIntersectRay2(p0, p1, p2, point, dir[0], dist))
			{
				if (0 <= dist) bits[0] = true;
				if (dist <= 0) bits[1] = true;
			}
			if (triangleIntersectRay2(p0, p1, p2, point, dir[1], dist))
			{
				if (0 <= dist) bits[2] = true;
				if (dist <= 0) bits[3] = true;
			}
			if (triangleIntersectRay2(p0, p1, p2, point, dir[2], dist))
			{
				if (0 <= dist) bits[4] = true;
				if (dist <= 0) bits[5] = true;
			}
		}
		return bits[0] && bits[1] && bits[2] && bits[3] && bits[4] && bits[5] ? true : false;
	}
	
	static Anchor[] createList(Mqo.Material[] material, Mqo.Object anchorObject)
	in
	{
		assert(isAnchor(anchorObject));
	}
	body
	{
		alias Vector[] VectorList;
		alias char[] MaterialName;
		VectorList[MaterialName] triangleMap;
		foreach (Mqo.Face face; anchorObject.face)
		{
			triangleMap[material[face.M].name] ~= createTriangle(anchorObject, face);
		}
		
		Anchor[] result;
		foreach (char[] a; triangleMap.keys)
		{
			result ~= new Anchor(triangleMap[a], a);
		}
		return result;
	}
	
	static Vector[] createTriangle(Mqo.Object obj, Mqo.Face face)
	{
		const uint[] indexList3 = [ 0, 1, 2 ];
		const uint[] indexList4 = [ 0, 1, 2,  0, 2, 3 ];
		uint[] indexList = face.vlength == 3 ? indexList3 : indexList4;
		Vector[] result;
		foreach (uint a; indexList) result ~= toVector(obj.vertex[face.V[a]]);
		return result;
	}
	
	private static Vector toVector(Mqo.Vector3 a) { return Vector.create(a.x, a.y, a.z); }
	
	
	unittest
	{
		assert("hoge" == extractRelationObjectName("hoge|hoge"));
		assert("aaa" == extractRelationObjectName("abc|aaa"));
		assert("abc" != extractRelationObjectName("abc|aaa"));
	}
	static char[] extractRelationalObjectName(Mqo.Object anchorObject)
	in
	{
		assert(isAnchor(anchorObject));
	}
	body
	{
		return extractRelationObjectName(anchorObject.name);
	}
	private static char[] extractRelationObjectName(char[] name)
	{
		char[][] a = split(name, "|");
		assert(2 == a.length);
		return a[1];
	}
}
