﻿module nanami.blackboard;
private import nanami.mqo;
private import nanami.mkm;
private import nanami.skeleton;
private import hdk.sdlimage;
private import std.utf;
private import hdk.math;
private import std.stream;
private import std.cstream;

interface KnowledgeSource
{
	bit condition();
	void opCall();
}

class Blackboard
{
	char[] mqoFileName, mkmFileName, mqoDirectory;
	bit flatvertex;
	
	void[] result;
	
	Mqo mqo;
	Mkm mkm;
	Skeleton skeleton;
	
	alias char[] string;
	string[][] textureFileName;
	
	Texture[][] texture;
	VertexBuffer[][] vertexBuffer;
	
	int opApply(int delegate(inout VertexBuffer vb, inout uint x, inout uint y) dg)
	{
		for (uint y = 0; y < vertexBuffer.length; y++)
		{
			for (uint x = 0; x < vertexBuffer[y].length; x++)
			{
				if (!vertexBuffer[y][x]) continue;
				int result = dg(vertexBuffer[y][x], x, y);
				if (result != 0) return result;
			}
		}
		return 0;
	}
	
	MotionMatrix[] motionMatrix;
}

class Texture
{
	uint[] rgba;
	uint width, height;
	invariant { assert(rgba.length == width * height); }
	
	this(char[] textureFileName)
	{
		rgba = readImage(toUTF32(textureFileName), width, height);
	}
	
	this(uint[] rgba, uint width, uint height)
	{
		this.rgba = rgba;
		this.width = width;
		this.height = height;
	}
	
	static Texture createNullTexture()
	{
		const uint[] rgba = [ uint.max ];
		return new Texture(rgba, 1, 1);
	}
	
	static Texture lerp(Texture previous, Texture next, float zero_one)
	{
		uint width = max(previous.width, next.width);
		uint height = max(previous.height, next.height);
		uint[] rgba = new uint[width * height];
		for (int y = 0; y < height; y++)
		{
			float fy = cast(float)y / cast(float)height;
			for (int x = 0; x < width; x++)
			{
				float fx = cast(float)x / cast(float)width;
				rgba[y * width + x] = lerp(previous[fx, fy], next[fx, fy], zero_one);
			}
		}
		return new Texture(rgba, width, height);
	}
	
	private static uint max(uint a, uint b) { return a >= b ? a : b; }
	
	private uint opIndex(float x, float y)
	in
	{
		assert(0.0 <= x && x <= 1.0);
		assert(0.0 <= y && y <= 1.0);
	}
	body
	{
		uint x2 = cast(uint)(x * width);
		uint y2 = cast(uint)(y * height);
		return rgba[y2 * width + x2];
	}
	
	private static uint lerp(uint pixelA, uint pixelB, float s)
	{
		ubyte[] result = new ubyte[4];
		ubyte[] ubaA = toUbyteArray(pixelA);
		ubyte[] ubaB = toUbyteArray(pixelB);
		result[0] = lerp(ubaA[0], ubaB[0], s);
		result[1] = lerp(ubaA[1], ubaB[1], s);
		result[2] = lerp(ubaA[2], ubaB[2], s);
		result[3] = lerp(ubaA[3], ubaB[3], s);
		return toUint(result);
	}
	
	private static ubyte lerp(ubyte a, ubyte b, float s)
	{
		return cast(ubyte)(a + (b - a) * s);
	}
	
	unittest
	{
		ubyte[] rgbaPixel = new ubyte[4];
		rgbaPixel[0] = 10;
		rgbaPixel[1] = 1;
		rgbaPixel[2] = 2;
		rgbaPixel[3] = 3;
		assert(rgbaPixel[0] == toUbyteArray(toUint(rgbaPixel))[0]);
		assert(rgbaPixel[1] == toUbyteArray(toUint(rgbaPixel))[1]);
		assert(rgbaPixel[2] == toUbyteArray(toUint(rgbaPixel))[2]);
		assert(rgbaPixel[3] == toUbyteArray(toUint(rgbaPixel))[3]);
	}
	
	private static ubyte[] toUbyteArray(uint a)
	out (result)
	{
		assert(4 == result.length);
	}
	body
	{
		ubyte[] result = new ubyte[4];
		result[3] = cast(ubyte)(a >> 24);
		result[2] = cast(ubyte)(a << 8 >> 24);
		result[1] = cast(ubyte)(a << 16 >> 24);
		result[0] = cast(ubyte)(a << 24 >> 24);
		return result;
	}
		
	private static uint toUint(ubyte[] rgbaPixel)
	in
	{
		assert(4 == rgbaPixel.length);
	}
	body
	{
		return (rgbaPixel[3] << 24) | (rgbaPixel[2] << 16)
			| (rgbaPixel[1] << 8) | rgbaPixel[0];
	}
}

class VertexBuffer
{
	uint sizeOfVertex = 0;
	float[] buffer;
	void add(float[] a) { buffer ~= a; }
	
	static VertexBuffer lerp(VertexBuffer previous, VertexBuffer next, float zero_one)
	in
	{
		assert(previous.sizeOfVertex == next.sizeOfVertex);
		assert(previous.buffer.length == next.buffer.length);
	}
	out (result)
	{
		assert(previous.buffer.length == result.buffer.length);
	}
	body
	{
		VertexBuffer result = new VertexBuffer();
		result.sizeOfVertex = previous.sizeOfVertex;
		result.buffer.length = previous.buffer.length;
		
		uint normalPosition = 7 * previous.vertexLength;
		uint texCoordPosition = 10 * previous.vertexLength;
		uint i = 0;
		
		while (i < normalPosition)
		{
			result.buffer[i] = .lerp(previous.buffer[i], next.buffer[i], zero_one);
			i++;
		}
		while (i < texCoordPosition)
		{
			float x = .lerp(previous.buffer[i], next.buffer[i], zero_one);
			float y = .lerp(previous.buffer[i + 1], next.buffer[i + 1], zero_one);
			float z = .lerp(previous.buffer[i + 2], next.buffer[i + 2], zero_one);
			Vector n = normalize(Vector.create(x, y, z));
			result.buffer[i] = n.x;
			result.buffer[i + 1] = n.y;
			result.buffer[i + 2] = n.z;
			i += 3;
		}
		while (i < previous.buffer.length)
		{
			result.buffer[i] = .lerp(previous.buffer[i], next.buffer[i], zero_one);
			i++;
		}
		
		return result;
	}
	
	uint vertexLength()
	{
		if (sizeOfVertex == 0) return 0;
		return buffer.length / sizeOfVertex;
	}
}

class MotionMatrix
{
	private Matrix[][] list;
	
	this(uint boneLength, uint frameLength)
	{
		list.length = frameLength;
		foreach (inout Matrix[] a; list) a.length = 3 * boneLength;
	}
	
	void setLrt(uint boneIndex, uint frameIndex,
		Matrix lookAt, Matrix rotation, Matrix translation)
	{
		list[frameIndex][3 * boneIndex + 0] = lookAt;
		list[frameIndex][3 * boneIndex + 1] = rotation;
		list[frameIndex][3 * boneIndex + 2] = translation;
	}
	
	Matrix getLookAt(uint t, uint boneIndex) { return list[t][3 * boneIndex + 0]; }
	Matrix getRotation(uint t, uint boneIndex) { return list[t][3 * boneIndex + 1]; }
	Matrix getTranslation(uint t, uint boneIndex) { return list[t][3 * boneIndex + 2]; }
	
	uint boneLength()
	{
		return list[0].length / 3;
	}
	
	uint frameLength()
	{
		return list.length;
	}
	
	void[] toData()
	out (result)
	{
		assert(result.length == uint.sizeof + float.sizeof * 16 * 3 * boneLength * frameLength);
	}
	body
	{
		MemoryStream result = new MemoryStream();
		result.write(frameLength);
		for (int fi = 0; fi < frameLength; fi++)
		{
			for (int bi = 0; bi < boneLength; bi++)
			{
				result.writeExact(&list[fi][3 * bi + 0], Matrix.sizeof);
				result.writeExact(&list[fi][3 * bi + 1], Matrix.sizeof);
				result.writeExact(&list[fi][3 * bi + 2], Matrix.sizeof);
			}
		}
		return result.data();
	}
	
	this(void[] data, bit trace = false)
	{
		if (trace) { dout.writeLine("MotionMatrix.this"); dout.flush(); }
		MemoryStream ms = new MemoryStream(cast(ubyte[])data);
		uint frameLength;
		ms.read(frameLength);
		list.length = frameLength;
		
		uint boneLength = (data.length - uint.sizeof) / Matrix.sizeof / frameLength / 3;
		if (trace)
		{
			dout.writefln("frameLength=%d boneLength=%d", frameLength, boneLength);
			dout.flush();
		}
		
		for (int f = 0; f < frameLength; f++)
		{
			list[f].length = 3 * boneLength;
			for (int b = 0; b < boneLength; b++)
			{
				Matrix l, r, t;
				ms.readExact(&l, Matrix.sizeof);
				ms.readExact(&r, Matrix.sizeof);
				ms.readExact(&t, Matrix.sizeof);
				setLrt(b, f, l, r, t);
			}
		}
	}
}
